#include <windows.h>
#include <math.h>

#include <stdio.h>
#include <string.h>
#include <conio.h>



typedef unsigned char uint8_t;
typedef short int int16_t;
typedef unsigned short int uint16_t;
typedef unsigned int uint32_t;


#define MYBAUDRATE CBR_19200
#define N_BYTES 10	//inclusive start and stop byte
#define STARTBYTE 237 
#define STOPBYTE 77


int receive_packet( HANDLE hComPort, uint8_t* status_packet );



void usage()
{
	fprintf( stdout, "Usage: printvoltage_ser comport\n" );
}

int main(int argc, char *argv[])
{
	void *dummy = NULL;
    char PortStr[6], 
		 PortStrFull[_MAX_PATH], 
		 PortTest[6];

	DCB dcb;
    HANDLE hComPort;
	int iPacketIsValid;
	unsigned char status_packet[100];
	BOOL bReturn;

	if (argc!=2)
	{
		usage();
		return 0;
	}

    strcpy(PortStr, argv[1]);
    PortStr[5] = '\0';

    /* Test to see if asking for a valid com port */
    strcpy( PortTest, PortStr );
    PortTest[3] = '\0';
    if (strcmp(PortTest,"com")==0)
    {
        if ((PortStr[3]<'1') | (PortStr[3]>'8'))
		{
			fprintf( stdout, "wrong input argument: comport\n" );
			return 0;
		}
    }
    else
	{
		fprintf( stdout, "wrong input argument: comport\n" );
		return 0;
	}

    PortStr[4] = '\0';
	strcpy(PortStrFull, "\\\\.\\");    /* use the full legal device name */
    strcat(PortStrFull, PortStr);

    hComPort = CreateFile( PortStrFull, 
          GENERIC_READ | GENERIC_WRITE,
          0,
          NULL,
          OPEN_EXISTING,
          0,
          NULL );

    if (hComPort == INVALID_HANDLE_VALUE)
	{
		fprintf( stdout, "open comport failed\n" );
		return 0;
	}

	// Omit the call to SetupComm to use the default queue sizes.
	// Get the current configuration.
	bReturn = GetCommState(hComPort, &dcb);
	if (!bReturn) 
	{
		fprintf( stdout, "GetCommState failed\n" );
		CloseHandle(hComPort);
		return 0;
	}

	// Fill in the DCB: baud=9600, 8 data bits, no parity, 1 stop bit. 
	dcb.BaudRate = MYBAUDRATE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	bReturn = SetCommState(hComPort, &dcb);
	if (!bReturn) 
	{
		fprintf( stdout, "SetCommState failed\n" );
		CloseHandle(hComPort);
		return 0;
	}


	fprintf( stdout, ";cycle, voltage, stopcounter\n" );
	fprintf( stdout, ";---------------------------\n" );

	while(!_kbhit())
	{

		iPacketIsValid = receive_packet( hComPort, status_packet );
				
		if (iPacketIsValid>0)
		{
			if (status_packet[6+1] == 81)
			{
				int16_t* x = (int16_t*)(status_packet+1);
				fprintf( stdout, "%i, %i, %i\n", x[1], x[0], x[2] );
			}
		}

		Sleep(100);

	}
	

	CloseHandle(hComPort);

	return 0;
}




int receive_packet( HANDLE hComPort, uint8_t* status_packet )
{
	BOOL bResult;
	DWORD nNumberOfBytesToRead;
	DWORD NumberOfBytesRead;

	int iPacketReceived = 0;
	int iPacketIsValid = 0;

	while ( iPacketReceived==0 )
	{
		nNumberOfBytesToRead = 1;
		bResult = ReadFile(hComPort, status_packet, nNumberOfBytesToRead, &NumberOfBytesRead, NULL);
		if ( (!bResult) || (nNumberOfBytesToRead!=NumberOfBytesRead) ) 
			return -1;

		if ( status_packet[0] == STARTBYTE )
		{
			nNumberOfBytesToRead = N_BYTES-1;
			bResult = ReadFile(hComPort, status_packet+1, nNumberOfBytesToRead, &NumberOfBytesRead, NULL);
			if ( (!bResult) || (nNumberOfBytesToRead!=NumberOfBytesRead) )
				return -1;

			iPacketReceived = 1;

			iPacketIsValid=0;
			if (status_packet[N_BYTES-1]==STOPBYTE)
			{
				int n;

				unsigned char sumbyte = 0;
				for ( n=1; n<(N_BYTES-2); n++ )
					sumbyte = (sumbyte + status_packet[n]);
				if (sumbyte==status_packet[N_BYTES-2]) 
					iPacketIsValid = 1;
			}
		}
	}

	return iPacketIsValid;
}


