;******************************************************************************
;*                                                                            *
;*  Firmware TRXCharger                                                       *
;*                                                                            *
;*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              *
;*  Project start: 01.08.2004                                                 *
;*  Last change  : 06.11.2005                                                 *
;*                                                                            *
;******************************************************************************
;*                                                                            *
;*  This program is free software; you can redistribute it and/or modify      *
;*  it under the terms of the GNU General Public License as published by      *
;*  the Free Software Foundation; either version 2 of the License, or         *
;*  (at your option) any later version.                                       *
;*                                                                            *
;*  This program is distributed in the hope that it will be useful,           *
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
;*  GNU General Public License for more details.                              *
;*                                                                            *
;*  You should have received a copy of the GNU General Public License         *
;*  along with this program; if not, write to the Free Software               *
;*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
;*                                                                            *
;******************************************************************************

;------------------- calibration values ------------------------------------------------------------------

.EQU  I_charge_factor_EEPROM = 17407;14100	
	;"charge current" = I_charge_code * (I_charge_factor/256) * ( U_ref_adc/(20*1024*R_shunt) )
	;	I_charge_factor NOT LARGER THAN (1024*256/15) !!!
	;	e.g. R_shunt=0.02Ohm, U_ref_adc=2.56V, I_charge_code=1, I_charge_factor=8192
	;	==> "Charge Current" = 200mA
.EQU  T_discharge_factor_EEPROM = 8*256		
	;"peak discharge current" = I_charge_code * T_discharge_factor/(8*256)*U_ref_adc*R2/(R1+R2)*10us/L
	;	e.g. L=100uH, R1=33k, R2=10k, U_ref_adc=2.56V, I_charge_code=1
	;	==> "peak discharge current" = 1A

;-----------------------------------------------------------------------------------------------------------------------

.EQU I_CHARGE_FACTOR_H_MAX = 68		; ~1024/15
.EQU SLOW_CHARGE_DIVISOR = 10		; consider a full cycle of about 1-second-charge and e.g. 10ms-pause
.EQU SLOW_CHARGE_CYCLES = 45818	    ; ( 14 hours ) * ( 3600 seconds/hour )
.EQU T_charge_cycletime = 31250	    ;charge cycle time = T_charge / ( 512/16MHz )
.EQU CALIBRATION_CHARGE_CODE = 10


;enable this to get data while charging
#define DEBUG 
;#define DEBUG_PARALLELPORT 		;else, serial-port is used for output
;#define INVERT_TXD		;inversion means, you can directly connect the TXD-pin to RS232-RXD

.include "tn26def.inc"

;***** Konstante
.equ SPL = $3d		; the same as SP
.equ RAMBEGIN = $60

.EQU  U_akku_on = 5					;0.1V/5V*(R2/(R1+R2))*1024     
.EQU  U_akku_min_default = 150
.EQU  U_akku_max_default = 1020		;

.EQU  WAITTIME_DISCHARGE_TO_MEASURE = 5
.EQU  WAITTIME_CHARGE_TO_DISCHARGE = 1
.EQU  WAITTIME_SLOW_CHARGE_TO_MEASURE = 10
.EQU  BLINKING_TIME_SLOW_CHARGE = (170*32)
.EQU  BLINKING_TIME_FAST_CHARGE = (40*32)
.EQU  BLINKING_TIME_FINISH = (1500*32)
.EQU  N_ADC_CYCLES_GET_VOLTAGE = 32

;PORTB
.EQU PIN_LED0 = 		0
.EQU PIN_LED1 = 		1
.EQU PIN_LED2 = 		2
.EQU PIN_LED3 = 		3
.EQU PIN_NFET = 		4
.EQU PIN_PFET = 		5
.EQU PIN_BUTTON = 		6

;PORTA
.EQU PIN_LED4 = 		0
.EQU PIN_IMINUS = 		1
.EQU PIN_IPLUS = 		2
.EQU PIN_AREF =			3
.EQU PIN_VCELL =		4
.EQU PIN_VCELL_SWITCH =	5
.EQU PIN_V0 = 			7


.EQU ADMUX_charge_current 	= 0b11000000 + 14	;REF = internal+capacitor, MUX = 01110 = (ADC2-ADC1)*20
.EQU ADMUX_voltage_check 	= 0b11000000 + 3	;REF = internal+capacitor, MUX = 00011 = ADC3
.EQU ADMUX_voltage_measure_AREF 	= 0b11000000 + 3	;REF = internal+capacitor, MUX = 00011 = ADC3

;serial output pin
.EQU TXD_PORT = PORTA
.EQU TXD_PIN = 6

;***** Variables (Register)
.DEF TIM1_state	 = R0
.DEF TIM1_AL 	 = R1
.DEF TIM1_AH 	 = R2
.DEF temp_int	 = R5

.DEF TIM1_I_l	 = R6
.DEF TIM1_I_h 	 = R7
.DEF TIM1_time_l = R8
.DEF TIM1_time_h = R9

.DEF DL			 = R10
.DEF DH			 = R11
.DEF EL			 = R12
.DEF EH			 = R13
.DEF one		 = R14
.DEF zero		 = R15

.DEF TIM1_dI_0 	= R22
.DEF TIM1_dI_1 	= R23
.DEF TIM1_dI_2  = R4
.DEF TIM1_dI0_0 = R24
.DEF TIM1_dI0_1 = R25
.DEF TIM1_dI0_2 = R3

.DEF AL         = r16	; my general purpose 16bit-register A, B, C
.DEF AH         = r17
.DEF BL         = r18
.DEF BH         = r19
.DEF CL			 = r20
.DEF CH			 = r21

.DEF temp       = R26

;***** Global Variables (SRAM)
.DSEG
.ORG RAMBEGIN 
U_akku_min_l:	.BYTE 1
U_akku_min_h:	.BYTE 1
U_akku_max_l:	.BYTE 1
U_akku_max_h:	.BYTE 1
I_charge_f:		.BYTE 1
I_charge_l:		.BYTE 1
I_charge_h:		.BYTE 1
I_charge_ctrl_l:	.BYTE 1
I_charge_ctrl_h:	.BYTE 1
T_charge_l:			.BYTE 1
T_charge_h:			.BYTE 1
T_discharge:		.BYTE 1
T_discharge_scaled:	.BYTE 1
I_charge_code:		.BYTE 1
Reflex_Charge__ADMUX:	.BYTE 1
TIM1_AL_init:		.BYTE 1
TIM1_AH_init:		.BYTE 1
threshold_counter:	.BYTE 1
td:					.BYTE 2
t_stop:				.BYTE 2
peak_enable_signal: .BYTE 2
peak_prediction_enable:	.BYTE 1
first_run:			.BYTE 1
delta_t_int:		.BYTE 2
LED_blinking_time_ms_low:	.BYTE 1
LED_blinking_time_ms_high:	.BYTE 1
ddrb_last:	.BYTE 1
status_packet_index:	.BYTE 1
status_packet:	.BYTE BYTES_PER_PACKET

.EQU  BYTES_PER_PACKET = 10


;***** EEPROM Variables
.ESEG
.ORG 0 
I_charge_factor:
.dw I_charge_factor_EEPROM
T_discharge_factor:	
.dw T_discharge_factor_EEPROM
I_charge_code_last: 
.db 1, 0 

;********************************************************************
;* Interrupt table                                                  *
;********************************************************************
.CSEG
.ORG 0				
	rjmp POWERON ; Reset handler
	rjmp EXT_INT0 ; IRQ0 handler
	rjmp PIN_CHANGE ; Pin change handler
	rjmp TIM1_CMP1A ; Timer1 compare match 1A
	rjmp TIM1_CMP1B ; Timer1 compare match 1B
	rjmp TIM1_OVF ; Timer1 overflow handler
	rjmp TIM0_OVF ; Timer0 overflow handler
	rjmp USI_STRT ; USI Start handler
	rjmp USI_OVF ; USI Overflow handler
;	rjmp EE_RDY ; EEPROM Ready handler
;	rjmp ANA_COMP ; Analog Comparator handler
;	rjmp ADC_COMPLETE ; ADC Conversion Handler


;********************************************************************
;* Init program                                                     *
;********************************************************************
POWERON:
	ser AL
	sts first_run, AL

RESET:
Init:
	;***** disable all interrupts
	cli

	;***** set stackpointer
	ldi	AL, RAMEND
	out	SPL, AL

	;***** set constants
	clr zero
	clr one
	inc one

	;***** Ports und Register initialisieren
	;PORTB
	;LED0..3 an, N/P-FET aus, Button auf GND, Pullup bei Button-Pin aktivieren
	ldi AL, (0<<PIN_LED0)|(0<<PIN_LED1)|(0<<PIN_LED2)|(0<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(1<<PIN_BUTTON)
    out PORTB, AL
	;LED-Pins und FET-Pins sind Ausgnge, Button ist Eingang
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(1<<PIN_NFET)|(1<<PIN_PFET)|(0<<PIN_BUTTON) 
    out DDRB, AL

	;PORTA
	;LED4 an, sonst alles aus / Pullup deaktivieren
	ldi AL, (0<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out PORTA, AL
	;LED4_PIN ist Ausgang, sonst alles Eingnge
	ldi AL, (1<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out DDRA, AL

#ifdef DEBUG
	#ifndef DEBUG_PARALLELPORT
	;enable serial DEBUG output, we dont disable it because of RS232 protocoll
    sbi (TXD_PORT-1), TXD_PIN		;DDR = PORT-1
	#ifdef INVERT_TXD
	cbi TXD_PORT, TXD_PIN
	#else
	sbi TXD_PORT, TXD_PIN
	#endif
	#endif
#endif

	;init ADC
	ldi AL, 0b10000111	;enable ADC, ADC_clk = clk/128
	out ADCSR, AL			

	;***** Settings
	;last charge current
	rcall load_I_charge_code; AL, AH is overridden!!!
	;set Umin
	ldi AL, LOW(U_akku_min_default)
	sts U_akku_min_l, AL
	ldi AH, HIGH(U_akku_min_default)
	sts U_akku_min_h, AH
	;set Umax
	ldi AL, LOW(U_akku_max_default)
	sts U_akku_max_l, AL
	ldi AH, HIGH(U_akku_max_default)
	sts U_akku_max_h, AH
	;set Cycle-Time
	ldi AL, LOW(T_charge_cycletime)
	sts T_charge_l, AL
	ldi AH, HIGH(T_charge_cycletime)
	sts T_charge_h, AH
	;just to check if value in EEPROM are reliable
	rcall get_I_charge_factor__BHBL

	;**** Interrupt-Handling
	out GIMSK, zero	;disable ext0 and pin-change interrupt
	out TIMSK, zero	;disable timer0/1 interrupt
	out MCUSR, zero	;clear all int-flags
	ser AL
	out GIFR, AL	;clear ext0 and pin-change int-flag (by writing logic one)
	;falling edge triggers int0
	in AL, MCUCR
	ori AL, 0b00000010
	out MCUCR, AL
	sei		; enable interrupts

	;go into testmode?
	;check if button is hold down for 3 seconds after poweron
	lds AL, first_run
	tst AL
	breq testmode_no
	rcall is_button_on_long
	breq testmode_no
	rjmp Testmode
	testmode_no:

	;**** wait some time 
	ldi AL, 250		;wait 250ms
	rcall wait_ms
	in AH, PORTB	;all LEDs off
	ori AH, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)
	out PORTB, AH
	sbi PORTA, PIN_LED4
	rcall wait_ms	;wait 250ms

	;**** wait until button released
	rcall wait_until_button_off
	sts first_run, zero

;********************************************************************
;* Main program                                                     *
;********************************************************************
Normal:
	; state-LED off
	sbi PORTA, PIN_LED4
	;display I_charge_code on LEDs
	rcall plot_charge_LEDs	; AL, AH is overridden!!!
	clr CL
Normal__loop:
	; check if accu is inserted
	rcall is_Voltage_on
	brne Pause
	; check if button is pressed
	rcall is_button_on
	brne Normal__increase_current
	; should we print voltage/current status?
	subi CL, -1
	brne Normal__loop
#ifdef DEBUG
	rcall print_Voltage		;caution: AL, AH is overridden!!!
#endif
	rjmp Normal__loop

Normal__increase_current:
	lds AL, I_charge_code
	mov BH, AL	;save value
	inc AL
	cpi AL, 16
	brlt Normal__increase_current_weiter
	ldi AL, 1
	Normal__increase_current_weiter:
	sts I_charge_code, AL
	rcall plot_charge_LEDs	; AL, AH is overridden!!!
	;check if button is hold down for 3 seconds
	rcall is_button_on_long
	breq Normal
	sts I_charge_code, BH
	rcall plot_charge_LEDs	; AL, AH is overridden!!!
	rcall save_I_charge_code; AL, AH is overridden!!!
	rjmp Laden_slow

Pause:
	; state-LED on
	cbi PORTA, PIN_LED4
	clr CL
Pause__loop:
	; check if accu is inserted
	rcall is_Voltage_on
	breq Normal
	; check if button is pressed
	rcall is_button_on
	brne Laden
	;should we print voltage/current status?
	subi CL, -16
	brne Pause__loop
#ifdef DEBUG
	rcall print_Voltage		;caution: AL, AH is overridden!!!
#endif
	rjmp Pause__loop

Laden:
	;save settings in eeprom
	rcall save_I_charge_code; AL, AH is overridden!!!
Laden_slow_or_fast:
	rcall is_button_on_long
	breq Laden_fast
	rjmp Laden_slow

Laden_fast:
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	lds AL, I_charge_code
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_ctrl_l, BH
	sts I_charge_ctrl_h, CL
	;set LED blinking time
	ldi AL, LOW(BLINKING_TIME_FAST_CHARGE)
	sts LED_blinking_time_ms_low, AL
	ldi AL, HIGH(BLINKING_TIME_FAST_CHARGE)
	sts LED_blinking_time_ms_high, AL
	;calculate and  set discharge current
	rcall get_T_discharge_factor__BHBL	; AL, AH is overridden!
	lds AL, I_charge_code
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts T_discharge, BH

	rcall enable_int0_button
	rcall Reflex_Charge;  ;caution: AL,AH,BL,BH,CL,CH,temp is overridden!!!

	rjmp Laden_Ende

Laden_slow:
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	mov AH, BH
	mov AL, BL
	ldi BL, SLOW_CHARGE_DIVISOR
	ldi BH, 0
	rcall div16u	;[AH,AL] rem [CH,CL] = [AH,AL] / [BH,BL] 
	lds BL, I_charge_code
	clr BH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_l, BH
	sts I_charge_h, CL
	;set LED blinking time
	ldi AL, LOW(BLINKING_TIME_SLOW_CHARGE) 
	sts LED_blinking_time_ms_low, AL
	ldi AL, HIGH(BLINKING_TIME_SLOW_CHARGE)
	sts LED_blinking_time_ms_high, AL
	
	rcall Timer_Charge ;caution: AL,BL,BH is overridden!!!

Laden_Ende:
	rcall set_state_LED_charge_sucess
Laden_Ende_loop:
	; wait until accu is removed
	rcall is_Voltage_on
	brne Laden_Ende_loop
	rjmp RESET


enable_int0_button:
	;enable int0, so button hit can abort charging
	ser AL			;clear int flag (by writing a logic "1")
	out GIFR, AL	
	out MCUSR, zero	;clear all int-flags
	in AL, GIMSK	;enable extern interrupt
	ori AL, 0b01000000
	out GIMSK, AL
	ret

set_state_LED_charge_sucess:
	clr TIM1_state		;state=idle
	rcall start_Timer1
	;enable LED slow blinking
	ldi AL, LOW(BLINKING_TIME_FINISH)	; 1s slow blinking
	ldi AH, HIGH(BLINKING_TIME_FINISH)
	rcall start_LED_blinking
	ret

Testmode:
	;set LEDs 
	sbi PORTA, PIN_LED4
	ldi AL, CALIBRATION_CHARGE_CODE
	sts I_charge_code, AL
	rcall plot_charge_LEDs	; AL, AH is overridden!
	;wait until button released
	rcall wait_until_button_off
	;set LED blinking time
	ldi AL, LOW(BLINKING_TIME_FAST_CHARGE)
	ldi AH, HIGH(BLINKING_TIME_FAST_CHARGE)
	rcall start_LED_blinking
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	ldi AL, CALIBRATION_CHARGE_CODE
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	mov TIM1_I_l, BH
	mov TIM1_I_h, CL
	;calculate and  set discharge current
	rcall get_T_discharge_factor__BHBL	; AL, AH is overridden!
	ldi AL, CALIBRATION_CHARGE_CODE
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts T_discharge, BH
	;preset charge cycle time
	ldi BL, 127
	mov TIM1_time_h, BL
	;set ADC
	ldi AL, ADMUX_charge_current
	out ADMUX, AL	
	sbi ADCSR, ADSC	;start first conversion, important for int-routine!!!
	;enable charging:
	ldi AL, 0b00000001
	or TIM1_state, AL
	;start Timer1
	rcall start_Timer1

	;charge until button pressed. 
Testmode_charge_loop:
	mov TIM1_time_h, BL; reset timeout for infinite charging
	rcall is_button_on
	breq Testmode_charge_loop
	mov TIM1_time_h, zero
	Testmode_charge_loop2:	;wait until charging has finished!
	sbrc TIM1_state, 0
	rjmp Testmode_charge_loop2

	rcall wait_until_button_off
	in DL, PORTB
Testmode_discharge_loop:
	;all charge-LEDs off
	in AL, PORTB
	ori AL, 0b00001111
	out PORTB, AL
	;set discharge time
	rcall set_T_discharge_scaled	;AL, AH, BL, BH, CL, CH is overridden!!!
	lds AL, T_discharge_scaled			; time in 2*10us for nfet-on, nfet-off
	;selected charge-LEDs on
	out PORTB, DL
	;discharge
	rcall discharge	
	rcall is_button_on
	breq Testmode_discharge_loop

	rjmp RESET


;***********************************************************
;* Interrupt functions                                     *
;***********************************************************

;external interrupt
EXT_INT0:
;	cli			; disable all interrupts
	rjmp RESET	;we reset stack-pointer, so rjmp is OK here

; Timer1 overflow handler
TIM1_OVF: 
	in temp_int, SREG
	sbrc TIM1_state, 0			;nfet-off, pfet-on (charge, increase current)
	rjmp TIM1_nfet_off_pfet_on
	rjmp TIM1_end
;nfet-off, pfet-on (charge)
TIM1_nfet_off_pfet_on:
	cbi PORTB, PIN_NFET	;switch off N-FET
	rcall wait_500ns	;wait 0.5us
	sbi PORTB, PIN_PFET	;switch on P-FET
TIM1_end:
	out SREG, temp_int
	reti


; Timer1 compare match 1A
TIM1_CMP1A: 
	in temp_int, SREG
	push temp
	push AL
	push AH

	;do something, in dependancy of the actual state
	sbrc TIM1_state, 0			;pfet-off, nfet-on (charge, dedrease current)
	rcall TIM1_CMP1A_charge
	sbrc TIM1_state, 2			;led-on-off
	rcall TIM1_CMP1A_led_on_off

TIM1_CMP1A_end:
	pop AH
	pop AL
	pop temp
	out SREG, temp_int
	reti

TIM1_CMP1A_charge:
	sbis PORTB, PIN_PFET		;check if PFET is on
	ret
	cbi PORTB, PIN_PFET	;switch off P-FET
	rcall wait_500ns	;wait 1us
	rcall wait_500ns
	sbi PORTB, PIN_NFET	;switch on N-FET
	; count cycles
	sub TIM1_time_l, one
	sbc TIM1_time_h, zero
	brne TIM1_CMP1A_charge__continue
	; stop charging if timeout
	cbi PORTB, PIN_NFET	;switch off N-FET
	sbi PORTA, PIN_LED4 ;state LED off
	clr TIM1_state	;disable charging
	rcall stop_Timer1
	ret
TIM1_CMP1A_charge__continue:
	;check if any ADC-conversion is completed
	sbis ADCSR, ADIF	; caution! ADIF-bit is automatically cleared 
	ret
	in TIM1_dI_0, ADCL	;get ADC-values
	in TIM1_dI_1, ADCH
	clr TIM1_dI_2
	;start new conversion
	sbi ADCSR, ADSC		; ADIF-bit is automatically cleared 
	;calculate control difference
	sub TIM1_dI_0, TIM1_I_l
	sbc TIM1_dI_1, TIM1_I_h
	sbc TIM1_dI_2, zero
	;integrate
	sub TIM1_dI0_0, TIM1_dI_0
	sbc TIM1_dI0_1, TIM1_dI_1
	sbc TIM1_dI0_2, TIM1_dI_2
	;limit integral
.EQU t_on_min = 2
.EQU t_on_max = 192

	cpi TIM1_dI0_1, t_on_min
	cpc TIM1_dI0_2, zero
	brlt TIM1_CMP1A_charge__too_low

	cpi TIM1_dI0_1, t_on_max
	cpc TIM1_dI0_2, zero
	brge TIM1_CMP1A_charge__too_big
	; set new output = I-part divided by 256
	out OCR1A, TIM1_dI0_1	;set output-compare-1A register
	ret
TIM1_CMP1A_charge__too_low:
	ldi temp, t_on_min
	out OCR1A, temp	;set output-compare-1A register
	clr TIM1_dI0_0
	ldi TIM1_dI0_1, t_on_min
	clr TIM1_dI0_2
	ret
TIM1_CMP1A_charge__too_big:
	ldi temp, t_on_max
	out OCR1A, temp	;set output-compare-1A register
	clr TIM1_dI0_0
	ldi TIM1_dI0_1, t_on_max
	clr TIM1_dI0_2
	ret

;led-on-off
TIM1_CMP1A_led_on_off:			
	sub TIM1_AL, one
	sbc TIM1_AH, zero
	breq TIM1_CMP1A_led_on_off2
	ret
TIM1_CMP1A_led_on_off2:
	lds TIM1_AL, TIM1_AL_init	; reload counter
	lds TIM1_AH, TIM1_AH_init
	sbis PORTA, PIN_LED4
	rjmp TIM1_CMP1A_led_off
	cbi PORTA, PIN_LED4
	ret
TIM1_CMP1A_led_off:
	sbi PORTA, PIN_LED4
	ret


#ifdef DEBUG
#ifdef DEBUG_PARALLELPORT
USI_OVF: ; USI Overflow handler
	in temp_int, SREG
	push AL
	sbis PINB, 2	; check if SCK is low, so we can change data register
	rjmp USI_OVF_new_data
	ldi AL, 0b11101111; clear int flags and set counter to 15, so next call, SCK is low
	USI_OVF_end:
	pop AL
	out SREG, temp_int
	reti

USI_OVF_new_data:
	lds AL, status_packet_index
	cpi AL, (BYTES_PER_PACKET-1)
	brge USI_OVF_no_new_data
	inc AL
	sts status_packet_index, AL
	USI_OVF_no_new_data:
	push ZL
	ldi ZL, status_packet
	add ZL, AL
	ld AL, Z
	pop ZL
	out USIDR, AL		;set new data
	sbi USISR, USIOIF	;clear int flag
	rjmp USI_OVF_end
#else
USI_OVF:	
#endif
#else
USI_OVF:
#endif


;***********************************************************
;* functions                                               *
;***********************************************************

;Adress in AL, result in AL
;AL is overridden!
read_EEPROM:
	sbic EECR, EEWE		;wait until ready
	rjmp read_EEPROM
	out EEAR, AL
	sbi EECR, EERE
	in AL, EEDR
	ret
	
get_I_charge_factor__BHBL:
	ldi AL, I_charge_factor
	rcall read_EEPROM
	mov BL, AL
	ldi AL, I_charge_factor+1
	rcall read_EEPROM
	mov BH, AL
	cpi AL, I_CHARGE_FACTOR_H_MAX	; check reliability: it has to be smaller than ~1024/15
	brlo get_I_charge_factor__BHBL_return
	rjmp Error
	get_I_charge_factor__BHBL_return:
	ret

get_T_discharge_factor__BHBL:
	ldi AL, T_discharge_factor
	rcall read_EEPROM
	mov BL, AL
	ldi AL, T_discharge_factor+1
	rcall read_EEPROM
	mov BH, AL
	ret

;load last charge-current-setting
load_I_charge_code:	; AL, AH is overridden!!!
	ldi AL, I_charge_code_last
	rcall read_EEPROM
	cpi AL, 1
	brlt load_I_charge_code__default
	cpi AL, 16
	brlt load_I_charge_code__continue
	load_I_charge_code__default:
	ldi AL, 1
	load_I_charge_code__continue:
	sts I_charge_code, AL
	ret

;save settings for charge-current in EEPROM
save_I_charge_code:	; AL, AH is overridden!!!	
	lds AL, I_charge_code
	save_I_charge_code__loop1:
	sbic EECR, EEWE
	rjmp save_I_charge_code__loop1	;wait until pending EEPROM-writing
	ldi BL, I_charge_code_last
	out EEAR, BL						;set adress and data
	out EEDR, AL
	cli	;disable interrupts for save EEPROM-writing
	sbi EECR, EEMWE					;MasterWriteEnable
	sbi EECR, EEWE					;WriteEnable
	save_I_charge_code__loop2:
	sbic EECR, EEWE
	rjmp save_I_charge_code__loop2	;wait until write is complete
	sei	;enable interrupts
	ret

plot_charge_LEDs:	; AL, AH is overridden!!!
	in AL, PORTB			
	andi AL, 0b11110000
	lds AH, I_charge_code
	com AH					;0=="LED on"
	andi AH, 0b00001111
	or AL, AH
	out PORTB, AL
	ret


update_statistics_dummy:
	;compensate time for "update_statistics"
	push AL
	ldi AL, 26
	update_statistics_dummy_loop:
	dec AL
	brne update_statistics_dummy_loop
	pop AL
	ret

update_statistics:
;input: [BH,BL] = t
	push AL
	push AH
	push BL
	push BH
	push CL
	push CH

	;get last predicted value td
	lds DL, td
	lds DH, td+1

	;delta_t = t - td
	sub BL, DL		;calculate difference
	sbc BH, DH

	;delta_t_int = delta_t_int + delta_t
	lds AL, delta_t_int
	lds AH, delta_t_int+1
	add AL, BL
	adc AH, BH
	sts delta_t_int, AL
	sts delta_t_int+1, AH

	;delta_t_int*k_I, hers 1/4
	subi AL, LOW(-2)	;rounding offset
	sbci AH, HIGH(-2)
	asr AH
	ror AL
	asr AH
	ror AL

	;(DELTA_T_THRESHOLD > delta_t) ? enable peak prediction : disable peak prediction
	ldi CL, LOW(-DELTA_T_THRESHOLD)
	ldi CH, HIGH(-DELTA_T_THRESHOLD)
	sub CL, BL
	sbc CH, BH
	sts peak_enable_signal, CL
	sts peak_enable_signal+1, CH
	brlt update_stat_enable_peak_not
	sts peak_prediction_enable, one
	update_stat_enable_peak_not:

	;delta_t*k_P, here 3/4 = 1-1/4: 
	mov EL, BL
	mov EH, BH
	asr EH	;/4
	ror EL
	asr EH
	ror EL
	sub BL, EL
	sbc BH, EH

	;td = td + delta_t*k_P + delta_t_int*k_I
	add BL, DL
	adc BH, DH
	add BL, AL
	adc BH, AH
	sts td, BL
	sts td+1, BH

	;calculate new t_stop = td + T_STOP_PLUS + delta_t_int*k_I/2
	asr AH
	ror AL
	add BL, AL
	adc BH, AH
	subi BL, LOW(-T_STOP_PLUS)	;time offset
	sbci BH, HIGH(-T_STOP_PLUS)
	sts t_stop, BL
	sts t_stop+1, BH

	pop CH
	pop CL
	pop BH
	pop BL
	pop AH
	pop AL
	ret

update_statistics_init:	;AL is overridden!!!
	sts peak_prediction_enable, zero
	ldi AL, LOW( DELTA_T_INT_INIT )
	ldi AH, HIGH( DELTA_T_INT_INIT )
	sts delta_t_int, AL
	sts delta_t_int+1, AH
	sts td, zero
	sts td+1, zero
	ret

is_negative_slope:
;input: [BH,BL] = t
;output: AL
	lds AL, peak_prediction_enable
	tst AL	
	breq is_negative_slope__no
	lds AL, t_stop
	lds AH, t_stop+1
	sub AL, BL; subtract actual time
	sbc AH, BH
	clr AL
	brcc is_negative_slope__no
	inc AL
	is_negative_slope__no:
	ret


set_T_discharge_scaled:	;AL, AH, BL, BH, CL, CH is overridden!!!
	sbi DDRA, PIN_VCELL_SWITCH	;enable voltage divider
	cbi PORTA, PIN_VCELL_SWITCH
	ldi AL, 50	;wait to stabilize voltage input
	rcall wait_ms
	ldi AL, ADMUX_voltage_measure_AREF
	rcall get_voltage
	rcall divide_AHAL_by_16
	mov BL, AL
	mov BH, AH
	lds AH, T_discharge	; *256
	clr AL
	rcall div16u	;[AH,AL] rem [CH,CL] = [AH,AL] / [BH,BL] 
	lds BL, T_discharge
	cp BL, AL	;new value has to be smaller!
	brsh set_T_discharge_scaled__cont
	mov AL, BL
	set_T_discharge_scaled__cont:
	sts T_discharge_scaled, AL
	ret


init_ADMUX:	;AL, AH is overridden!!!
.EQU U_accu_threshold = 700*32		;maximum (empty) accu ADC-value in corresponding range
	;0V .. ~2.5V - accu no load, 1 cell
	cbi DDRA, PIN_VCELL_SWITCH	;disable voltage divider
	cbi PORTA, PIN_VCELL_SWITCH	;disable pull-up-resistor
	ldi AL, 50	;wait to stabilize Voltage input
	rcall wait_ms
	ldi AL, ADMUX_voltage_measure_AREF
	sts Reflex_Charge__ADMUX, AL
	rcall get_voltage
	cpi AL, LOW(U_accu_threshold)
	ldi AL, HIGH(U_accu_threshold)
	cpc AH, AL
	brlt init_ADMUX__ret
	;0V .. ~10V - accu no load, 2..6 cells
	sbi DDRA, PIN_VCELL_SWITCH	;enable voltage divider
	ldi AL, ADMUX_voltage_measure_AREF
	sts Reflex_Charge__ADMUX, AL
	init_ADMUX__ret:
	ret


Reflex_Charge: ;caution: AL,AH,BL,BH,CL,CH,temp is overridden!!!
.EQU T_startup = 60		;no of cycles, no dU-check is made, has to be positive!
.EQU dU_threshold = 2		;dU_threshold_in_V = dU_threshold*U_ref_adc/1024; should be bigger than one (hysterese)!
.EQU threshold_counter_init = 5
.EQU OCR1A_init = 64
.EQU DELTA_T_THRESHOLD = 10	;has to be positive!
.EQU T_STOP_PLUS = 10
.EQU DELTA_T_INT_INIT = -30*4
.EQU Reflex_Charge__maxtime = (3600*3)	;~3h

	;init Variables
	ldi AL, OCR1A_init	;set default on time
	out OCR1A, AL
;	clr TIM1_dI0_0 ;precision is not needed here
	mov TIM1_dI0_1, AL	;preset Integral
	clr TIM1_dI0_2
	rcall update_statistics_init	;AL is overridden!!!

	;set initial charge current and precharge some time without stopping, to get a reliable voltage for ADMUX setting
	lds AL, I_charge_ctrl_l
	lds AH, I_charge_ctrl_h
	rcall divide_AHAL_by_16
;	sts I_charge_f, zero ;precision is not needed here
	sts I_charge_l, AL
	sts I_charge_h, AH

	;precharge some time
	rcall charge		;1second
	ldi AL, 30
	rcall wait_ms

	;set discharge time
	rcall set_T_discharge_scaled	;AL, AH, BL, BH, CL, CH is overridden!!!

	;set proper voltage divider, ADC-reference for voltage measurement
	rcall init_ADMUX	;AL, AH is overridden!!!

	clr CL	;last Voltage
	clr CH
	clr BL	;time counter
	clr BH
Reflex_Charge__loop:
	;is time out?
	add BL, one
	adc BH, zero
	cpi BL, LOW(Reflex_Charge__maxtime)
	ldi temp, HIGH(Reflex_Charge__maxtime)
	cpc BH, temp
	brcs Reflex_Charge__no_timeout
	rjmp Reflex_Charge__finish
	Reflex_Charge__no_timeout:

;slowly ( 1-exp(-t/tau) ) increase current up to normal value: x(n) = x(n-1) + ( c - x(n-1) + 126)/128
	mov DH, zero
	lds AL, I_charge_ctrl_l
	lds AH, I_charge_ctrl_h
	lds DL, I_charge_f
	lds EL, I_charge_l
	lds EH, I_charge_h
	sub DH, DL
	sbc AL, EL
	sbc AH, EH
	rcall divide_AHALDH_by_128_offset
	add DL, DH
	adc EL, AL	; 
	adc EH, AH
	sts I_charge_f, DL
	sts I_charge_l, EL
	sts I_charge_h, EH
	
	;wait xx ms
	ldi AL, WAITTIME_DISCHARGE_TO_MEASURE
	rcall wait_ms

	;Voltage Measurement, about 4ms
	lds AL, Reflex_Charge__ADMUX
	rcall get_voltage

#ifdef DEBUG
	sts status_packet+1, AL
	sts status_packet+2, AH
	sts status_packet+3, BL
	sts status_packet+4, BH
	lds EL, t_stop
	lds EH, t_stop+1
	sub EL, BL
	sbc EH, BH
	lds DL, peak_prediction_enable	;just to mark "enable peak prediction" add 256 if disabled
	tst DL	
	brne output_peak_prediction_enable
	lds EL, peak_enable_signal
	lds EH, peak_enable_signal+1
	output_peak_prediction_enable:
	sts status_packet+5, EL
	sts status_packet+6, EH
	ldi temp, PACKET_ID_ADC_REFLEX	;packet ID
	sts status_packet+7, temp
	rcall release_packet
#endif

	rcall divide_AHAL_by_32

	;voltage too low?
	lds temp, U_akku_min_l
	cp AL, temp
	lds temp, U_akku_min_h
	cpc AH, temp
	brge Reflex_Charge__no_low_voltage
	rjmp Reflex_Charge__low_voltage
	Reflex_Charge__no_low_voltage:	;voltage too high?
	lds temp, U_akku_max_l
	cp AL, temp
	lds temp, U_akku_max_h
	cpc AH, temp
	brlt Reflex_Charge__no_high_voltage
	rjmp Reflex_Charge__high_voltage
	Reflex_Charge__no_high_voltage:

	;normal mode
Reflex_Charge__normal_mode:
	;-dU criteria: 
	cp  CL, AL	; U_max - U_measure
	cpc CH, AH
	brge Reflex_Charge__check_dU
	rcall update_statistics		;we have a new maximum!
	mov CL, AL	
	mov CH, AH
	ldi temp, threshold_counter_init
	sts threshold_counter, temp
	rjmp Reflex_Charge__dU_is_positive

Reflex_Charge__check_dU:
	rcall update_statistics_dummy
	sub AL, CL	;calculate difference, U_measure - U_max
	sbc AH, CH
	cpi AL, LOW(1-dU_threshold)	;dU lower than threshold?
	ldi temp, HIGH(1-dU_threshold)
	cpc AH, temp
	brge Reflex_Charge__dU_is_positive
	lds temp, threshold_counter
	dec temp
	sts threshold_counter, temp
	;brne Reflex_Charge__do_charging
	brne Reflex_Charge__dU_is_positive
	rjmp Reflex_Charge__stop

Reflex_Charge__dU_is_positive:
	rcall is_negative_slope
	breq Reflex_Charge__do_charging

Reflex_Charge__stop:
	;startup-phase?	
	cpi BL, LOW(T_startup)
	ldi temp, HIGH(T_startup)
	cpc BH, temp
	brcc Reflex_Charge__finish	;if carry is clear, compare is positive

Reflex_Charge__do_charging:
	;charge
	rcall charge

	;wait xx ms
	ldi AL, WAITTIME_CHARGE_TO_DISCHARGE
	rcall wait_ms

	;discharge
	lds AL, T_discharge_scaled			; time in 2*10us for nfet-on, nfet-off
	;ramp I 
	rcall discharge	
	;ramp II 
	rcall discharge
	;ramp III 
	rcall discharge	

rjmp Reflex_Charge__loop

Reflex_Charge__low_voltage:
Reflex_Charge__high_voltage:
Reflex_Charge__finish:
	ret

	divide_AHAL_by_16:
	ldi temp, 4
	rjmp divide_AHAL_by_32_loop

	divide_AHAL_by_32:
	;divide by 32
	subi AL, LOW(-16)	;round offset ( 16/32 = 0.5 )
	sbci AH, HIGH(-16)
	ldi temp, 5	;=log2(N_ADC_CYCLES_GET_VOLTAGE)
	divide_AHAL_by_32_loop:
	lsr AH
	ror AL
	dec temp
	brne divide_AHAL_by_32_loop
	ret

	divide_AHALDH_by_128_offset:	; x = (x/2 + 63)/64
	lsr AH				; /2
	ror AL
	ror DH
	ldi temp, LOW(-63)
	sub DH, temp
	sbci AL, HIGH(-63)	;offset +32 
	sbci AH, HIGH(-63)
	ldi temp, 6			; /64
	divide_AHALDH_by_128_loop:
	lsr AH
	ror AL
	ror DH
	dec temp
	brne divide_AHALDH_by_128_loop
	ret


Timer_Charge: ;caution: AL,BL,BH is overridden!!!
	;init Variables
	ldi BL, LOW(SLOW_CHARGE_CYCLES)	;set loop counter 
	ldi BH, HIGH(SLOW_CHARGE_CYCLES)

Timer_Charge__button_on_loop:
	rcall is_button_on
	breq Timer_Charge__cont
	rcall charge
	;wait xxx ms
	ldi AL, WAITTIME_SLOW_CHARGE_TO_MEASURE
	rcall wait_ms
	rjmp Timer_Charge__button_on_loop

Timer_Charge__cont:
;	ldi AL, 50
;	rcall wait_ms
	rcall enable_int0_button
Timer_Charge__loop:
	;is time out?
	sub BL, one
	sbc BH, zero
	brcs Timer_Charge__finish

	;charge 1 second
	rcall charge

	;wait xxx ms
	ldi AL, WAITTIME_SLOW_CHARGE_TO_MEASURE
	rcall wait_ms

	;check if Accu is inserted furthermore
	rcall is_Voltage_on
	brne Timer_Charge__loop
	Timer_Charge__finish:
	ret


;arguments are global variables: I_charge_l, I_charge_h, T_charge_l, T_charge_h
charge:
	push AL
	push AH
	lds TIM1_I_l, I_charge_l
	lds TIM1_I_h, I_charge_h
	lds TIM1_time_l, T_charge_l
	lds TIM1_time_h, T_charge_h

	;set ADC
	ldi AL, ADMUX_charge_current
	out ADMUX, AL	
			
	;do dummy conversion, and also wait to stabilize MUX input
	sbi ADCSR, ADSC
charge__waitloop0:
	sbis ADCSR, ADIF
	rjmp charge__waitloop0

	;start initial conversion
	sbi ADCSR, ADSC
charge__waitloop1:
	in AL, ADCSR
	sbrs AL, ADIF
	rjmp charge__waitloop1

	;set timer-state: enable charging
	ldi AL, 0b00000001
	mov TIM1_state, AL

	;enable LED fast blinking
	lds AL, LED_blinking_time_ms_low
	lds AH, LED_blinking_time_ms_high
	rcall start_LED_blinking

	;start Timer1
	rcall start_Timer1

charge__waitloop2:
	sbrc TIM1_state, 0
	rjmp charge__waitloop2
	;stop LED-blinking
	rcall stop_LED_blinking

	pop AH
	pop AL
	ret


;input argument in AL: ADMUX,
;output in AH:AL: ADC-result (32*11bit=15bit)
get_voltage:
	push temp

	;get actual Voltage
	out ADMUX, AL	
	ldi AL, 1
	rcall wait_ms
			
	;do dummy conversion
	sbi ADCSR, ADSC
Voltage_Measurement__waitloop1:
	sbis ADCSR, ADIF
	rjmp Voltage_Measurement__waitloop1

	;do N_ADC_CYCLES_GET_VOLTAGE conversions, takes about ???ms 
	clr AL
	clr AH
	ldi temp, N_ADC_CYCLES_GET_VOLTAGE
Voltage_Measurement__waitloop2:
	sbi ADCSR, ADSC
Voltage_Measurement__waitloop3:
	sbis ADCSR, ADIF
	rjmp Voltage_Measurement__waitloop3
	in EL, ADCL	;get ADC-values
	in EH, ADCH
	add AL, EL	;accumulate
	adc AH, EH
	dec temp
	brne Voltage_Measurement__waitloop2

	pop temp
	ret


discharge:		
	mov EL, AL
	tst EL
	breq discharge_end
	sbic PORTB, PIN_PFET	;check if PFET is off
	rjmp Error
	sbi PORTB, PIN_NFET	;switch on NFET
discharge_loop1:
	rcall wait_10us
	dec EL
	brne discharge_loop1
	cbi PORTB, PIN_NFET	;switch off NFET
	mov EL, AL
discharge_loop2:
	rcall wait_10us
	dec EL
	brne discharge_loop2
discharge_end:
	ret		


start_Timer1:
	push AL
	;stop timer
	out TCCR1B, zero
	; clear timer interrupt-flag
	out TIFR, zero
	;reset counter
	out TCNT1, zero
	;enable overflow and output-compare-1A interrupt
	in AL, TIMSK
	ori AL, 0b01000100
	out TIMSK, AL
	; set prescaler and start timer
	ldi AL, 0b00000010	; CK/2
	out TCCR1B, AL
	pop AL
	ret


stop_Timer1:
	push AL
	; stop timer
	out TCCR1B, zero
	;clear timer interrupt-flag
	out TIFR, zero
	;disable overflow and output-compare-1A interrupt, 
	in AL, TIMSK
	andi AL, 0b10111011
	out TIMSK, AL
	pop AL
	ret

start_LED_blinking:
	push AL
	sts TIM1_AL_init, AL
	sts TIM1_AH_init, AH
	; set cycle time
	mov TIM1_AL, AL
	mov TIM1_AH, AH
	; initial setting: LED=on
	cbi PORTA, PIN_LED4
	;set state of int-routine: led on/off
	ldi AL, 0b00000100
	or TIM1_state, AL
	pop AL
	ret

stop_LED_blinking:
	push AL
	;set state of int-routine:
	;state = idle	
	ldi AL, 0b11111011
	and TIM1_state, AL
	; switch off LED
	sbi PORTA, PIN_LED4
	pop AL
	ret


TIM0_OVF:
PIN_CHANGE: ; Pin change handler
TIM1_CMP1B: ; Timer1 compare match 1B
USI_STRT: ; USI Start handler
EE_RDY: ; EEPROM Ready handler
ANA_COMP: ; Analog Comparator handler
ADC_COMPLETE: ; ADC Conversion Handler

Error:
	; disable all Interrupts
	cli
	Error__loop:
	;PORTB zurcksetzen
	;LED0..3 an, N/P-FET aus, Button auf GND
	ldi AL, (0<<PIN_LED0)|(0<<PIN_LED1)|(0<<PIN_LED2)|(0<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(0<<PIN_BUTTON)
    out PORTB, AL
	;LED-Pins und FET-Pins sind Ausgnge, Button ist Eingang
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(1<<PIN_NFET)|(1<<PIN_PFET)|(0<<PIN_BUTTON) 
    out DDRB, AL
	;PORTA zurcksetzen
	;LED4 an, sonst alles aus / Pullup deaktivieren
	ldi AL, (0<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH) 
    out PORTA, AL
	;LED4_PIN ist Ausgang, sonst alles Eingnge
	ldi AL, (1<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out DDRA, AL
	ldi AL, 125
	rcall wait_ms
	;all LEDs off, Pullup bei Button-Pin aktiviert lassen
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(0<<PIN_BUTTON)
    out PORTB, AL
	sbi PORTA, PIN_LED4
	ldi AL, 125
	rcall wait_ms
	rjmp Error__loop

;***********************************************************
;* wait_ms                                                 *
;***********************************************************
;input arguments: AL = time in ms
wait_ms:
; assuming a clock frequency of 16 MHz ==> 1ms = 16000 cycles
; this is not exactly 1ms 
	push AL
	push AH
	ori AL, 0				; 
	breq wait_ms_end		; 
	wait_ms_loop1:
	ldi AH, 100				; 
	wait_ms_loop2:
	rcall wait_10us
	dec AH					; 
	brne wait_ms_loop2		; 
	dec AL					; 
	brne wait_ms_loop1		; 
	wait_ms_end:
	pop AH
	pop AL
	ret							

;***********************************************************
;* wait_10us                                              *
;***********************************************************
wait_10us:
	push AL
	ldi AL, 13
	wait_10us_loop:
	rcall wait_500ns
	dec AL
	brne wait_10us_loop
	pop AL	; same as 4 nops
	push AL
	nop
	nop
	pop AL
	ret	

;***********************************************************
;* wait_500ns                                              *
;***********************************************************
wait_500ns:
; assuming a clock frequency of 16 MHz ==> 500ns = 8 cycles
	nop
	ret	

;***********************************************************
;* is_button_on/off                                        *
;***********************************************************

is_button_on:
	is_button_on__yes_maybe:
	clr AL
	is_button_on__yes_count:
	sbic PINB, PIN_BUTTON
	rjmp is_button_on__no_maybe
	inc AL
	breq is_button_on__yes
	rcall wait_10us
	rjmp is_button_on__yes_count
	is_button_on__no_maybe:
	clr AL
	is_button_on__no_count:
	sbis PINB, PIN_BUTTON
	rjmp is_button_on__yes_maybe
	inc AL
	breq is_button_on__no
	rcall wait_10us
	rjmp is_button_on__no_count
	is_button_on__yes:
	inc AL
	is_button_on__no:
	ret

;check if button is hold for a longer time, here 3s
is_button_on_long:
	push BL
	ldi BL, 30
	is_button_on_long_maybe:
	rcall is_button_on
	breq is_button_on_long__no
	ldi AL, 100
	rcall wait_ms
	dec BL
	brne is_button_on_long_maybe
	is_button_on_long__yes:
	inc BL
	ldi AL, 1
	is_button_on_long__no:
	pop BL
	ret

;wait_until_button_on:
;	rcall is_button_on
;	breq wait_until_button_on
;	ret

wait_until_button_off:
	rcall is_button_on
	brne wait_until_button_off
	ret

.EQU Voltage_on_threshold_h = 2
is_Voltage_on:
	push AH
	push BL
	push temp

	;set voltage divider, ADC-reference for voltage measurement
	sbi DDRA, PIN_VCELL_SWITCH	;enable voltage divider
	cbi PORTA, PIN_VCELL_SWITCH	;disable pull-up-resistor

	ldi BL, (256-20)
	is_Voltage_on__yes_count:
	ldi AL, ADMUX_voltage_measure_AREF
	rcall get_voltage
	;voltage too low?
	ldi temp, LOW(U_akku_on*N_ADC_CYCLES_GET_VOLTAGE)	; 32 because of 32 ADC-measurements in get Voltage
	cp AL, temp
	ldi temp, HIGH(U_akku_on*N_ADC_CYCLES_GET_VOLTAGE)
	cpc AH, temp

	brlt is_Voltage_on__no
	inc BL
	breq is_Voltage_on__yes
	rjmp is_Voltage_on__yes_count

	is_Voltage_on__yes:
	clr AL
	inc AL
	rjmp is_Voltage_on__return

	is_Voltage_on__no:
	clr AL
	is_Voltage_on__return:
	pop temp
	pop BL
	pop AH
	ret

#ifdef DEBUG
print_Voltage:	;caution: AL, AH  is overridden!!!
	;set voltage divider, ADC-reference for voltage measurement
	sbi DDRA, PIN_VCELL_SWITCH	;enable voltage divider
	cbi PORTA, PIN_VCELL_SWITCH	;disable pull-up-resistor
	ldi AL, ADMUX_voltage_measure_AREF
	rcall get_voltage
	sts status_packet+1, AL
	sts status_packet+2, AH
	ldi AL, ADMUX_charge_current
	rcall get_voltage
	sts status_packet+3, AL
	sts status_packet+4, AH
	ldi AL, PACKET_ID_PRINT_VOLTAGE
	sts status_packet+7, AL
	rcall release_packet
	ret
#endif

;***************************************************************************
;*
;* "mpy16u" - 16x16 Bit Unsigned Multiplication
;*
;* This subroutine multiplies the two 16-bit register variables 
;* mp16uH:mp16uL and mc16uH:mc16uL.
;* The result is placed in m16u3:m16u2:m16u1:m16u0.
;*  
;* Number of words	:14 + return
;* Number of cycles	:153 + return
;* Low registers used	:None
;* High registers used  :7 (mp16uL,mp16uH,mc16uL/m16u0,mc16uH/m16u1,m16u2,
;*                          m16u3,mcnt16u)	
;*
;***************************************************************************
;***** Subroutine Register Variables
.def	mc16uL	=r16		;multiplicand low byte
.def	mc16uH	=r17		;multiplicand high byte
.def	mp16uL	=r18		;multiplier low byte
.def	mp16uH	=r19		;multiplier high byte
.def	m16u0	=r18		;result byte 0 (LSB)
.def	m16u1	=r19		;result byte 1
.def	m16u2	=r20		;result byte 2
.def	m16u3	=r21		;result byte 3 (MSB)
.def	mcnt16u	=r26		;loop counter
;***** Code
mpy16u:	
	push r26
	clr	m16u3		;clear 2 highest bytes of result
	clr	m16u2
	ldi	mcnt16u,16	;init loop counter
	lsr	mp16uH
	ror	mp16uL
m16u_1:	brcc	noad8		;if bit 0 of multiplier set
	add	m16u2,mc16uL	;add multiplicand Low to byte 2 of res
	adc	m16u3,mc16uH	;add multiplicand high to byte 3 of res
noad8:	ror	m16u3		;shift right result byte 3
	ror	m16u2		;rotate right result byte 2
	ror	m16u1		;rotate result byte 1 and multiplier High
	ror	m16u0		;rotate result byte 0 and multiplier Low
	dec	mcnt16u		;decrement loop counter
	brne	m16u_1		;if not done, loop more
	pop r26
	ret



;***************************************************************************
;*
;* "div16u" - 16/16 Bit Unsigned Division
;*
;* This subroutine divides the two 16-bit numbers 
;* "dd8uH:dd8uL" (dividend) and "dv16uH:dv16uL" (divisor). 
;* The result is placed in "dres16uH:dres16uL" and the remainder in
;* "drem16uH:drem16uL".
;*  
;* Number of words	:19
;* Number of cycles	:235/251 (Min/Max)
;* Low registers used	:2 (drem16uL,drem16uH)
;* High registers used  :5 (dres16uL/dd16uL,dres16uH/dd16uH,dv16uL,dv16uH,
;*			    dcnt16u)
;*
;***************************************************************************

;***** Subroutine Register Variables

.def	drem16uL=r20
.def	drem16uH=r21
.def	dres16uL=r16
.def	dres16uH=r17
.def	dd16uL	=r16
.def	dd16uH	=r17
.def	dv16uL	=r18
.def	dv16uH	=r19
.def	dcnt16u	=r26

;***** Code

div16u:	
	push dcnt16u
	clr	drem16uL	;clear remainder Low byte
	sub	drem16uH,drem16uH;clear remainder High byte and carry
	ldi	dcnt16u,17	;init loop counter
d16u_1:	rol	dd16uL		;shift left dividend
	rol	dd16uH
	dec	dcnt16u		;decrement counter
	brne	d16u_2		;if done
	pop dcnt16u
	ret			;    return
d16u_2:	rol	drem16uL	;shift dividend into remainder
	rol	drem16uH
	sub	drem16uL,dv16uL	;remainder = remainder - divisor
	sbc	drem16uH,dv16uH	;
	brcc	d16u_3		;if result negative
	add	drem16uL,dv16uL	;    restore remainder
	adc	drem16uH,dv16uH
	clc			;    clear carry to be shifted into result
	rjmp	d16u_1		;else
d16u_3:	sec			;    set carry to be shifted into result
	rjmp	d16u_1
	

.EQU COMM_WAIT_TIME_MS = 5
.EQU PACKET_STARTBYTE = 237
.EQU PACKET_STOPBYTE = 77
.EQU PACKET_ID_STOP = 99
.EQU PACKET_ID_PRINT_VOLTAGE = 80
.EQU PACKET_ID_ADC_REFLEX = 81
.EQU PACKET_ID_SLOPE_REFLEX = 82

#ifdef DEBUG
#ifdef DEBUG_PARALLELPORT
;data transfer using the parallel programming adapter
;our simple data format: startbyte, b1, b2, b3, b4, b5, paritybyte, stopbyte
	release_packet:
	push AL
	push ZL
	out USICR, zero	;reset USI control register
	out USIDR, zero	;clear USI data register
	ldi AL, 0b11101100	;clear int-flags (by writing logic one), set counter to 12
	out USISR, AL
	in AL, DDRB			;save ddr portb 
	sts ddrb_last, AL
	andi AL, 0b11111010	; PB0, PB2 is input
	ori AL, 0b00000010	; PB1 is output
	out DDRB, AL 		; set port direction
	;prepare our packet frame
	ldi AL, PACKET_STARTBYTE
	sts status_packet+0, AL
	ldi AL, PACKET_STOPBYTE
	sts status_packet+(BYTES_PER_PACKET-1), AL
	ldi ZL, (status_packet+1)	; calculate sum (data only)
	ld EL, Z+
	ldi AL, (BYTES_PER_PACKET-4)
	output_data_loop:
	ld EH, Z+
	add EL, EH
	dec AL
	brne output_data_loop
	sts status_packet+(BYTES_PER_PACKET-2), EL
	ser AL						;reset index counter
	sts status_packet_index, AL
	;set USI control register: counter overflow enable, three wire mode, external/positive edge
	ldi AL, 0b01011000	
	out USICR, AL
	ldi AL, COMM_WAIT_TIME_MS
	rcall wait_ms
	out USICR, zero;reset USI control register
	lds EL, ddrb_last	;restore ddr portb 
	out DDRB, EL		; set port direction
	pop ZL
	pop AL
	ret
#else

;***************************************************************************
;*
;* "putchar"
;*
;* This subroutine transmits the byte stored in the "Txbyte" register
;* The number of stop bits used is set with the sb constant
;*
;* Number of words	:14 including return
;* Number of cycles	:Depens on bit rate
;* Low registers used	:None
;* High registers used	:2 (bitcnt,Txbyte)
;* Pointers used	:None
;*
;***************************************************************************
.equ		sb	=1		;Number of stop bits (1, 2, ...)
.equ		b = 75; 19200 bps		;b=(1/clk_data - 9/clk_CPU)/(500ns + 3/clk_CPU), 37,06 <==> 38400bps @ 16MHz  74,94 <==> 19200bps @ 16MHz,  150,70 <==> 9600bps @ 16MHz
.DEF bitcnt = r19
.DEF tmp = r20

putchar:	
	push bitcnt
	push AL	; byte to transmit
	push tmp

		ldi	bitcnt,9+sb	;1+8+sb (sb is # of stop bits)
		com	AL		;Invert everything
		sec			;Start bit

putchar0:	;    send a '0'					
		brcc	putchar1	;If carry set		1,2
		#ifdef INVERT_TXD
		sbi TXD_PORT, TXD_PIN					;2
		#else
		cbi TXD_PORT, TXD_PIN					
		#endif
		rjmp	putchar2	;else				3,4

putchar1:	;    send a '1'
		#ifdef INVERT_TXD
		cbi TXD_PORT, TXD_PIN					;3
		#else
		sbi TXD_PORT, TXD_PIN
		#endif
		nop										;4

putchar2:
UART_delay:	
		ldi	tmp,b							;	5
UART_delay1:
		rcall wait_500ns	;500ns
		dec	tmp				;1
		brne	UART_delay1	;2,3

		lsr	AL		;Get next bit				;6
		dec	bitcnt		;If not all bit sent	;7
		brne	putchar0	;   send next		;8,9
					;else
	pop tmp
	pop AL
	pop bitcnt
	ret			;return

	release_packet:
	push AL
	push ZL

	;prepare our packet frame
	ldi AL, PACKET_STARTBYTE
	sts status_packet+0, AL
	ldi AL, PACKET_STOPBYTE
	sts status_packet+(BYTES_PER_PACKET-1), AL
	ldi ZL, (status_packet+1)	; calculate sum (data only)
	ld EL, Z+
	ldi AL, (BYTES_PER_PACKET-4)
	output_data_loop:
	ld EH, Z+
	add EL, EH
	dec AL
	brne output_data_loop
	sts status_packet+(BYTES_PER_PACKET-2), EL
	
	ldi ZL, status_packet	
	putchar_data_loop:
	ld AL, Z+
	rcall putchar
	cpi ZL, (status_packet+BYTES_PER_PACKET)
	brne putchar_data_loop

	pop ZL
	pop AL
	ret
#endif

#else
	release_packet:
	ret
#endif
