;******************************************************************************
;*                                                                            *
;*  Firmware TRXCharger                                                       *
;*                                                                            *
;*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              *
;*  Project start: 01.08.2004                                                 *
;*  Last change  : 06.11.2005                                                 *
;*                                                                            *
;******************************************************************************
;*                                                                            *
;*  This program is free software; you can redistribute it and/or modify      *
;*  it under the terms of the GNU General Public License as published by      *
;*  the Free Software Foundation; either version 2 of the License, or         *
;*  (at your option) any later version.                                       *
;*                                                                            *
;*  This program is distributed in the hope that it will be useful,           *
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
;*  GNU General Public License for more details.                              *
;*                                                                            *
;*  You should have received a copy of the GNU General Public License         *
;*  along with this program; if not, write to the Free Software               *
;*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
;*                                                                            *
;******************************************************************************
;------------------- calibration values ------------------------------------------------------------------

.EQU  I_charge_factor_EEPROM = 14100	
	;"charge current" = I_charge_code * (I_charge_factor/256) * ( U_ref_adc/(20*1024*R_shunt) )
	;	I_charge_factor NOT LARGER THAN (1024*256/15) !!!
	;	e.g. R_shunt=0.02Ohm, U_ref_adc=2.56V, I_charge_code=1, I_charge_factor=8192
	;	==> "Charge Current" = 200mA
.EQU  T_discharge_factor_EEPROM = 8*256		
	;"peak discharge current" = I_charge_code * T_discharge_factor/(8*256)*U_ref_adc*R2/(R1+R2)*10us/L
	;	e.g. L=100uH, R1=33k, R2=10k, U_ref_adc=2.56V, I_charge_code=1
	;	==> "peak discharge current" = 1A

;-----------------------------------------------------------------------------------------------------------------------



.EQU I_CHARGE_FACTOR_H_MAX = 68

.include "tn26def.inc"

;***** Konstante
.equ SPL = $3d		; the same as SP
.equ RAMBEGIN = $60

;PORTB
.EQU PIN_LED0 = 	0
.EQU PIN_LED1 = 	1
.EQU PIN_LED2 = 	2
.EQU PIN_LED3 = 	3
.EQU PIN_NFET = 	4
.EQU PIN_PFET = 	5
.EQU PIN_BUTTON = 	6
.EQU PIN_FREE1 = 	7

;PORTA
.EQU PIN_LED4 = 	0
.EQU PIN_V0 = 		7
.EQU PIN_IPLUS = 	2
.EQU PIN_IMINUS = 	1
.EQU PIN_VCELL =	4
.EQU PIN_VCELL_SWITCH =	5
.EQU PIN_AREF =		3
.EQU PIN_FREE2 = 	6


 


.EQU ADMUX_charge_current 	= 0b11000000 + 14	;REF = internal+capacitor, MUX = 01110 = (ADC2-ADC1)*20
.EQU ADMUX_voltage_check 	= 0b11000000 + 3	;REF = internal+capacitor, MUX = 00011 = ADC3
.EQU ADMUX_voltage_measure_AREF 	= 0b11000000 + 3	;REF = internal+capacitor, MUX = 00011 = ADC3


;testmode output
.EQU PLOT16_PORT = PORTA
.EQU PLOT16_PIN = 7
.EQU TXD_PORT = PORTA
.EQU TXD_PIN = 7

;***** Variables (Register)
.DEF TIM1_state	 = R0
.DEF TIM1_AL 	 = R1
.DEF TIM1_AH 	 = R2
.DEF temp_int	 = R5

.DEF TIM1_I_l	 = R6
.DEF TIM1_I_h 	 = R7
.DEF TIM1_time_l = R8
.DEF TIM1_time_h = R9

.DEF DL			 = R10
.DEF DH			 = R11
.DEF EL			 = R12
.DEF EH			 = R13
.DEF one		 = R14
.DEF zero		 = R15

.DEF TIM1_dI_0 	= R22
.DEF TIM1_dI_1 	= R23
.DEF TIM1_dI_2  = R4
.DEF TIM1_dI0_0 = R24
.DEF TIM1_dI0_1 = R25
.DEF TIM1_dI0_2 = R3

.DEF AL         = r16	; my general purpose 16bit-register A, B, C
.DEF AH         = r17
.DEF BL         = r18
.DEF BH         = r19
.DEF CL			= r20
.DEF CH			= r21

.DEF temp       = R26

;***** Global Variables (SRAM)
.EQU  U_akku_min_l =	RAMBEGIN + 0x00	;0.9V/2.56V*1024
.EQU  U_akku_min_h =	RAMBEGIN + 0x01	;0.9V/2.56V*1024
.EQU  U_akku_max_l =	RAMBEGIN + 0x02	;1.8V/2.56*1024
.EQU  U_akku_max_h =	RAMBEGIN + 0x03	;1.8V/2.56*1024
.EQU  I_charge_l = 		RAMBEGIN + 0x04	;I = I_charge * ( R_shunt*20*1024/U_ref_adc )
.EQU  I_charge_h = 		RAMBEGIN + 0x05	;I = I_charge * ( R_shunt*20*1024/U_ref_adc )
.EQU  T_charge_l = 		RAMBEGIN + 0x06	;charge time = T_charge / ( 512/16MHz )
.EQU  T_charge_h = 		RAMBEGIN + 0x07	;charge time = T_charge / ( 512/16MHz )
.EQU  T_discharge = 	RAMBEGIN + 0x08	;time in 2*10us for nfet-on, nfet-off ;ramp I (shortcircuit with coil for T seconds: Imax = T*U_accu/L)
.EQU  T_discharge_scaled = RAMBEGIN + 0x09
.EQU  I_charge_code =	RAMBEGIN + 0x0A
.EQU  Reflex_Charge__ADMUX = RAMBEGIN + 0x0B
.EQU  TIM1_AL_init = RAMBEGIN + 0x0C
.EQU  TIM1_AH_init = RAMBEGIN + 0x0D
.EQU  threshold_counter = RAMBEGIN + 0x0E
.EQU  t_stop_plus = RAMBEGIN + 0x0F
.EQU  t_stop =			RAMBEGIN + 0x10	; 2 Byte
.EQU  dt_stop =			RAMBEGIN + 0x12	; 3 Byte
.EQU  t_last = 			RAMBEGIN + 0x15	; 2 Byte
.EQU  no_of_data_counter =  	RAMBEGIN + 0x17	; 1 Byte
.EQU  status_packet = RAMBEGIN + 24	; 8 bytes
.EQU  status_packet_index = RAMBEGIN + 32 ; 1 byte
.EQU  ddrb_last = RAMBEGIN + 33; 1 byte




;***** EEPROM Variables
.ESEG
.ORG 0 
I_charge_factor:
.dw I_charge_factor_EEPROM
T_discharge_factor:	
.dw T_discharge_factor_EEPROM
I_charge_code_last: 
.db 1, 0 


;********************************************************************
;* Interrupt table
;********************************************************************
.cseg
.org 0				
	rjmp RESET ; Reset handler
	rjmp EXT_INT0 ; IRQ0 handler
	rjmp PIN_CHANGE ; Pin change handler
	rjmp TIM1_CMP1A ; Timer1 compare match 1A
	rjmp TIM1_CMP1B ; Timer1 compare match 1B
	rjmp TIM1_OVF ; Timer1 overflow handler
	rjmp TIM0_OVF ; Timer0 overflow handler
	rjmp USI_STRT ; USI Start handler
	rjmp USI_OVF ; USI Overflow handler
	rjmp EE_RDY ; EEPROM Ready handler
	rjmp ANA_COMP ; Analog Comparator handler
	rjmp ADC_COMPLETE ; ADC Conversion Handler


;********************************************************************
;* Init program
;********************************************************************
RESET:
Init:
	;***** disable all interrupts
	cli

	;***** set stackpointer
	ldi	AL, RAMEND
	out	SPL, AL

	;***** set constants
	clr zero
	clr one
	inc one

	;***** Ports und Register initialisieren
	;PORTB
	;LED0..3 an, N/P-FET aus, Button auf GND
	ldi AL, (0<<PIN_LED0)|(0<<PIN_LED1)|(0<<PIN_LED2)|(0<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(0<<PIN_BUTTON)
    out PORTB, AL
	;LED-Pins und FET-Pins sind Ausgnge, Button ist Eingang
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(1<<PIN_NFET)|(1<<PIN_PFET)|(0<<PIN_BUTTON) 
    out DDRB, AL
	;Pullup bei Button-Pin aktivieren
	sbi PORTB, PIN_BUTTON

	;PORTA
	;LED4 an, sonst alles aus / Pullup deaktivieren
	ldi AL, (0<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out PORTA, AL
	;LED4_PIN ist Ausgang, sonst alles Eingnge
	ldi AL, (1<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out DDRA, AL

	;enable DEBUG output
;    sbi (TXD_PORT-1), TXD_PIN		;DDR = PORT-1
;	cbi TXD_PORT, TXD_PIN

	;init ADC
	ldi AL, 0b10000111	;enable ADC, ADC_clk = clk/128
	out ADCSR, AL			

	;**** Interrupt-Handling
	clr AL
	out GIMSK, AL	;disable ext0 and pin-change interrupt
	out TIMSK, AL	;disable timer0/1 interrupt
	out MCUSR, AL	;clear all int-flags
	ser AL
	out GIFR, AL	;clear ext0 and pin-change int-flag (by writing logic one)
	;falling edge triggers int0
	in AL, MCUCR
	ori AL, 0b00000010
	out MCUCR, AL
	; configure Watchdog-Timer
	;ldi tmp, 0b00001110		; ~1s @ VCC=5V
	;wdr
	;out wdtcr, tmp
	sei		; enable interrupts

	;init and start Timer1
	clr TIM1_state	;state=idle
	rcall start_Timer1


;********************************************************************
;* Main program                                                     *
;********************************************************************
START:


;********************************************************************
;* testmode program                                                 *
;********************************************************************


testmode:
	;*** (I) LED-test ***
	;disable blinking
	rcall stop_LED_blinking
	;switch on all LEDs
	in temp, PORTB
	andi temp, 0b11110000
	out PORTB, temp
	in temp, PORTA
	andi temp, 0b11111110
	out PORTA, temp

	;switch off on LED after another
	rcall wait_until_button_pressed
	sbi PORTB, PIN_LED3
	rcall wait_until_button_pressed
	sbi PORTB, PIN_LED2
	rcall wait_until_button_pressed
	sbi PORTB, PIN_LED1
	rcall wait_until_button_pressed
	sbi PORTB, PIN_LED0
	rcall wait_until_button_pressed
	sbi PORTA, PIN_LED4

	;*** (II) Port-test (Pins: PA2, PA4, PA6, PA7) *** 
	rcall wait_until_button_pressed

	ldi AL, LOW(1000*32)	; ; 1000 ms slow blinking, 1tick = 1/(clk/1024) = 1.064ms
	ldi AH, HIGH(1000*32)
	rcall start_LED_blinking
	;switch off LED 3..0
	in temp, PORTB
	ori temp, 0b00001111
	out PORTB, temp

	;PA2
	;1
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTA, 2	; high,
	sbi DDRA, 2		; output
	;2
	rcall wait_until_button_pressed
	dec temp				;2
	out PORTB, temp
	cbi PORTA, 2	; low
	;3
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi DDRA, 2		; input, high Z
	;PA4
	;4
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTA, 4	; high,
	sbi DDRA, 4		; output
	;5
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi PORTA, 4	; low
	;6
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi DDRA, 4		; input, high Z
	;PA6
	;7
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTA, 6	; high,
	sbi DDRA, 6		; output
	;8
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi PORTA, 6	; low
	;9
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi DDRA, 6		; input, high Z
	;PA7
	;10
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTA, 7	; high,
	sbi DDRA, 7		; output
	;11
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi PORTA, 7	; low
	;12
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	cbi DDRA, 7		; input, high Z

FET_test:
	;*** (III) FET-test ***
	rcall wait_until_button_pressed
	;switch off LED 3..0
	in temp, PORTB
	ori temp, 0b00001111
	out PORTB, temp

	;enable LED fast blinking
	ldi AL, LOW(100*32)	; 100 ms fast blinking
	ldi AH, HIGH(100*32)
	rcall start_LED_blinking

	;N-FET
	;1, on
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTB, PIN_NFET	; FET on
	;2, off
	rcall wait_until_button_pressed
	dec temp				;2
	out PORTB, temp
	cbi PORTB, PIN_NFET	; FET off

	;P-FET
	;3, on
	rcall wait_until_button_pressed
	dec temp
	out PORTB, temp
	sbi PORTB, PIN_PFET	; FET on
	;4, off
	rcall wait_until_button_pressed
	dec temp				;2
	out PORTB, temp
	cbi PORTB, PIN_PFET	; FET off


charge_test:
	;switch off LED 3..0
	in temp, PORTB
	ori temp, 0b00001111
	out PORTB, temp

	;enable LED fast blinking
	ldi AL, LOW(50*32)	;
	ldi AH, HIGH(50*32)
	rcall start_LED_blinking

	rcall wait_until_button_pressed
;charge constant current (low current)
;1) I = I_charge_min/10

	dec temp
	out PORTB, temp
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	mov AH, BH
	mov AL, BL
	ldi BL, 10
	ldi BH, 0
	rcall div16u	;[AH,AL] rem [CH,CL] = [AH,AL] / [BH,BL] 
	ldi BL, 1 ;I_charge_code
	clr BH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_l, BH
	sts I_charge_h, CL
	ldi AL, 127
	sts T_charge_l, AL
	sts T_charge_h, AL

	rcall charge
	rcall wait_until_button_pressed
	mov TIM1_time_l, zero
	mov TIM1_time_h, zero
	charge_test1_loop:
	sbrc TIM1_state, 0
	rjmp charge_test1_loop
	rcall stop_LED_blinking


;charge constant current (normal current)
;2) I = I_charge_min
	dec temp
	out PORTB, temp
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	ldi AL, 1 ;I_charge_code
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_l, BH
	sts I_charge_h, CL
	ldi AL, 127
	sts T_charge_l, AL
	sts T_charge_h, AL

	rcall charge
	rcall wait_until_button_pressed
	mov TIM1_time_l, zero
	mov TIM1_time_h, zero
	charge_test2_loop:
	sbrc TIM1_state, 0
	rjmp charge_test2_loop
	rcall stop_LED_blinking


;charge constant current (normal current)
;3) I = 10*I_charge_min
	dec temp
	out PORTB, temp
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	ldi AL, 10 ;I_charge_code
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_l, BH
	sts I_charge_h, CL
	ldi AL, 127
	sts T_charge_l, AL
	sts T_charge_h, AL

	rcall charge
	rcall wait_until_button_pressed
	mov TIM1_time_l, zero
	mov TIM1_time_h, zero
	charge_test3_loop:
	sbrc TIM1_state, 0
	rjmp charge_test3_loop
	rcall stop_LED_blinking


;charge constant current (normal current)
;4) I = 15*I_charge_min
	dec temp
	out PORTB, temp
	;calculate and set charge current
	rcall get_I_charge_factor__BHBL	; AL, AH is overridden!
	ldi AL, 15 ;I_charge_code
	clr AH
	rcall mpy16u	; [CH,CL,BH,BL] = [AH,AL] * [BH,BL]
	sts I_charge_l, BH
	sts I_charge_h, CL
	ldi AL, 127
	sts T_charge_l, AL
	sts T_charge_h, AL

	rcall charge
	rcall wait_until_button_pressed
	mov TIM1_time_l, zero
	mov TIM1_time_h, zero
	charge_test4_loop:
	sbrc TIM1_state, 0
	rjmp charge_test4_loop
	rcall stop_LED_blinking





	rjmp RESET



;***********************************************************
;* Interrupt functions                                     *
;***********************************************************

;external interrupt
EXT_INT0:
	cli			; disable all interrupts
	rjmp RESET


; Timer1 overflow handler
TIM1_OVF: 
	in temp_int, SREG
	sbrc TIM1_state, 0			;nfet-off, pfet-on (charge)
	rjmp TIM1_nfet_off_pfet_on
TIM1_end:
	out SREG, temp_int
	reti

;nfet-off, pfet-on (charge)
TIM1_nfet_off_pfet_on:
	cbi PORTB, PIN_NFET	;switch off N-FET
	rcall wait_500ns	;wait 0.5us
	sbi PORTB, PIN_PFET	;switch on P-FET
TIM1_end2:
	out SREG, temp_int
	reti


; Timer1 compare match 1A
TIM1_CMP1A: 
	in temp_int, SREG
	push temp
	push AL
	push AH

	;do something, in dependancy of the actual state
	sbrc TIM1_state, 0			;pfet-off, nfet-on (charge)
	rcall TIM1_CMP1A_charge
	sbrc TIM1_state, 2			;led-on-off
	rcall TIM1_CMP1A_led_on_off

TIM1_CMP1A_end:
	pop AH
	pop AL
	pop temp
	out SREG, temp_int
	reti

TIM1_CMP1A_charge:
	sbis PORTB, PIN_PFET		;check if PFET is on
	ret
	cbi PORTB, PIN_PFET	;switch off P-FET
	rcall wait_500ns	;wait 1us
	rcall wait_500ns
	sbi PORTB, PIN_NFET	;switch on N-FET
	; count cycles
	;sub TIM1_time_l, one
	sub TIM1_time_l, zero
	sbc TIM1_time_h, zero
	brne TIM1_CMP1A_charge__continue
	; stop charging if timeout
	cbi PORTB, PIN_NFET	;switch off N-FET
	rcall stop_Timer1
	clr TIM1_state	;disable charging
	ret
TIM1_CMP1A_charge__continue:
	;check if any ADC-conversion is completed
	sbis ADCSR, ADIF	; caution! ADIF-bit is automatically cleared 
	ret
	in TIM1_dI_0, ADCL	;get ADC-values
	in TIM1_dI_1, ADCH
	clr TIM1_dI_2
	;start new conversion
	sbi ADCSR, ADSC		; ADIF-bit is automatically cleared 
	;calculate control difference
	sub TIM1_dI_0, TIM1_I_l
	sbc TIM1_dI_1, TIM1_I_h
	sbc TIM1_dI_2, zero
	;integrate
	sub TIM1_dI0_0, TIM1_dI_0
	sbc TIM1_dI0_1, TIM1_dI_1
	sbc TIM1_dI0_2, TIM1_dI_2
	;limit integral
.EQU t_on_min = 10
.EQU t_on_max = 192

	cpi TIM1_dI0_1, t_on_min
	cpc TIM1_dI0_2, zero
	brlt TIM1_CMP1A_charge__too_low

	cpi TIM1_dI0_1, t_on_max
	cpc TIM1_dI0_2, zero
	brge TIM1_CMP1A_charge__too_big
	; set new output = I-part divided by 256
	out OCR1A, TIM1_dI0_1	;set output-compare-1A register
	ret
TIM1_CMP1A_charge__too_low:
	ldi temp, t_on_min
	out OCR1A, temp	;set output-compare-1A register
	clr TIM1_dI0_0
	ldi TIM1_dI0_1, t_on_min
	clr TIM1_dI0_2
	ret
TIM1_CMP1A_charge__too_big:
	ldi temp, t_on_max
	out OCR1A, temp	;set output-compare-1A register
	clr TIM1_dI0_0
	ldi TIM1_dI0_1, t_on_max
	clr TIM1_dI0_2
	ret

;led-on-off
TIM1_CMP1A_led_on_off:			
	sub TIM1_AL, one
	sbc TIM1_AH, zero
	breq TIM1_CMP1A_led_on_off2
	ret
TIM1_CMP1A_led_on_off2:
	lds TIM1_AL, TIM1_AL_init	; reload counter
	lds TIM1_AH, TIM1_AH_init
	sbis PORTA, PIN_LED4
	rjmp TIM1_CMP1A_led_off
	cbi PORTA, PIN_LED4
	ret
TIM1_CMP1A_led_off:
	sbi PORTA, PIN_LED4
	ret

TIM0_OVF:
PIN_CHANGE: ; Pin change handler
TIM1_CMP1B: ; Timer1 compare match 1B
USI_STRT: ; USI Start handler
USI_OVF: ; USI Overflow handler
EE_RDY: ; EEPROM Ready handler
ANA_COMP: ; Analog Comparator handler
ADC_COMPLETE: ; ADC Conversion Handler
rjmp RESET




;***********************************************************
;* functions                                               *
;***********************************************************





;Adress in AL, result in AL
;AL is overridden!
read_EEPROM:
	sbic EECR, EEWE		;wait until ready
	rjmp read_EEPROM
	out EEAR, AL
	sbi EECR, EERE
	in AL, EEDR
	ret
	
get_I_charge_factor__BHBL:
	ldi AL, I_charge_factor
	rcall read_EEPROM
	mov BL, AL
	ldi AL, I_charge_factor+1
	rcall read_EEPROM
	mov BH, AL
	cpi AL, I_CHARGE_FACTOR_H_MAX	; check reliability: it has to be smaller than 1024/15
	brlo get_I_charge_factor__BHBL_return
	rjmp Error
	get_I_charge_factor__BHBL_return:
	ret

get_T_discharge_factor__BHBL:
	ldi AL, T_discharge_factor
	rcall read_EEPROM
	mov BL, AL
	ldi AL, T_discharge_factor+1
	rcall read_EEPROM
	mov BH, AL
	ret




;arguments are global variables: I_charge_l, I_charge_h, T_charge_l, T_charge_h
charge:
	push AL
	push AH
	lds TIM1_I_l, I_charge_l
	lds TIM1_I_h, I_charge_h
	lds TIM1_time_l, T_charge_l
	lds TIM1_time_h, T_charge_h

	;set ADC
	ldi AL, ADMUX_charge_current
	out ADMUX, AL	
			
	;do dummy conversion, and also wait to stabilize MUX input
	sbi ADCSR, ADSC
charge__waitloop0:
	sbis ADCSR, ADIF
	rjmp charge__waitloop0

	;start initial conversion
	sbi ADCSR, ADSC
charge__waitloop1:
	in AL, ADCSR
	sbrs AL, ADIF
	rjmp charge__waitloop1

	;clear timer-state
	clr TIM1_state

	;enable LED fast blinking
	ldi AL, LOW(50*32)	; 50 ms fast blinking
	ldi AH, HIGH(50*32)
	rcall start_LED_blinking

	;enable charging:
	ldi AL, 0b00000001
	or TIM1_state, AL

	;start Timer1
	rcall start_Timer1

;charge__waitloop2:
;	sbrc TIM1_state, 0
;	rjmp charge__waitloop2
;	;stop LED-blinking
;	rcall stop_LED_blinking

	pop AH
	pop AL
	ret




start_Timer1:
	push AL
	;stop timer
	ldi AL, 0b00000000
	out TCCR1B, AL
	; clear timer interrupt-flag
	clr AL
	out TIFR, AL
	;reset counter
	out TCNT1, zero
	;enable overflow and output-compare-1A interrupt
	in AL, TIMSK
	ori AL, 0b01000100
	out TIMSK, AL
	; set prescaler and start timer
	ldi AL, 0b00000010	; CK/2
	out TCCR1B, AL
	pop AL
	ret


stop_Timer1:
	push AL
	; stop timer
	ldi AL, 0b00000000
	out TCCR1B, AL
	;clear timer interrupt-flag
	clr AL
	out TIFR, AL
	;disable overflow and output-compare-1A interrupt, 
	in AL, TIMSK
	andi AL, 0b10111011
	out TIMSK, AL
	pop AL
	ret

start_LED_blinking:
	push AL
	sts TIM1_AL_init, AL
	sts TIM1_AH_init, AH
	; set cycle time
	mov TIM1_AL, AL
	mov TIM1_AH, AH
	; initial setting: LED=on
	cbi PORTA, PIN_LED4
	;set state of int-routine: led on/off
	ldi AL, 0b00000100
	or TIM1_state, AL
	pop AL
	ret

stop_LED_blinking:
	push AL
	;set state of int-routine:
	;state = idle	
	ldi AL, 0b11111011
	and TIM1_state, AL
	; switch off LED
	sbi PORTA, PIN_LED4
	pop AL
	ret


Error:
	; disable all Interrupts
	cli
	Error__loop:
	;PORTB zurcksetzen
	;LED0..3 an, N/P-FET aus, Button auf GND
	ldi AL, (0<<PIN_LED0)|(0<<PIN_LED1)|(0<<PIN_LED2)|(0<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(0<<PIN_BUTTON)
    out PORTB, AL
	;LED-Pins und FET-Pins sind Ausgnge, Button ist Eingang
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(1<<PIN_NFET)|(1<<PIN_PFET)|(0<<PIN_BUTTON) 
    out DDRB, AL
	;PORTA zurcksetzen
	;LED4 an, sonst alles aus / Pullup deaktivieren
	ldi AL, (0<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH) 
    out PORTA, AL
	;LED4_PIN ist Ausgang, sonst alles Eingnge
	ldi AL, (1<<PIN_LED4) | (0<<PIN_V0) | (0<<PIN_IPLUS) | (0<<PIN_IMINUS) | (0<<PIN_VCELL) | (0<<PIN_VCELL_SWITCH)
    out DDRA, AL
	ldi AL, 125
	rcall wait_ms
	;all LEDs off, Pullup bei Button-Pin aktiviert lassen
	ldi AL, (1<<PIN_LED0)|(1<<PIN_LED1)|(1<<PIN_LED2)|(1<<PIN_LED3)|(0<<PIN_NFET)|(0<<PIN_PFET)|(0<<PIN_BUTTON)
    out PORTB, AL
	sbi PORTA, PIN_LED4
	ldi AL, 125
	rcall wait_ms
	rjmp Error__loop

;***********************************************************
;* wait_ms                                                 *
;***********************************************************
;input arguments: AL = time in ms
wait_ms:
; assuming a clock frequency of 16 MHz ==> 1ms = 16000 cycles
; this is not exactly 1ms 
	push AL
	push AH
	ori AL, 0				; 
	breq wait_ms_end		; 
	
	wait_ms_loop1:

	ldi AH, 21				; 
	wait_ms_loop2:

	clr EL
	wait_ms_loop3:
	dec EL					;1
	brne wait_ms_loop3		;2,3

	dec AH					; 
	brne wait_ms_loop2		; 

	dec AL					; 
	brne wait_ms_loop1		; 

	wait_ms_end:
	pop AH
	pop AL
	ret							


;***********************************************************
;* wait_100us                                              *
;***********************************************************
wait_100us:
	push AL
	ldi AL, 144
	
	wait_100us_loop:
	rcall wait_500ns
	dec AL
	brne wait_100us_loop
	pop AL	; same as 4 nops
	push AL
	nop
	pop AL
	ret	


;***********************************************************
;* wait_10us                                              *
;***********************************************************
wait_10us:
	push AL
	ldi AL, 13
	
	wait_10us_loop:
	rcall wait_500ns
	dec AL
	brne wait_10us_loop
	pop AL	; same as 4 nops
	push AL
	nop
	nop
	pop AL
	ret	

;***********************************************************
;* wait_500ns                                              *
;***********************************************************
wait_500ns:
; assuming a clock frequency of 16 MHz ==> 500ns = 8 cycles
	nop
	ret	

;***********************************************************
;* is_button_on/off                                        *
;***********************************************************

is_button_on:
	is_button_on__yes_maybe:
	clr AL
	is_button_on__yes_count:
	sbic PINB, PIN_BUTTON
	rjmp is_button_on__no_maybe
	inc AL
	breq is_button_on__yes
	rcall wait_10us
	rjmp is_button_on__yes_count
	is_button_on__no_maybe:
	clr AL
	is_button_on__no_count:
	sbis PINB, PIN_BUTTON
	rjmp is_button_on__yes_maybe
	inc AL
	breq is_button_on__no
	rcall wait_10us
	rjmp is_button_on__no_count
	is_button_on__yes:
	inc AL
	is_button_on__no:
	ret


wait_until_button_on:
	rcall is_button_on
	breq wait_until_button_on
	ret

wait_until_button_off:
	rcall is_button_on
	brne wait_until_button_off
	ret




;input argument in AL: ADMUX,
;output in AH:AL: ADC-result (32*11bit=15bit)
get_voltage:
	push temp
	push BL
	push BH

	;get actual Voltage
	out ADMUX, AL	
	ldi AL, 1
	rcall wait_ms
			
	;do dummy conversion
	sbi ADCSR, ADSC
Voltage_Measurement__waitloop1:
	sbis ADCSR, ADIF
	rjmp Voltage_Measurement__waitloop1

	;do 64 conversions, takes about 8ms 
	clr AL
	clr AH
	ldi temp, 32
Voltage_Measurement__waitloop2:
	sbi ADCSR, ADSC
Voltage_Measurement__waitloop3:
	sbis ADCSR, ADIF
	rjmp Voltage_Measurement__waitloop3
	in BL, ADCL	;get ADC-values
	in BH, ADCH
	add AL, BL	;accumulate
	adc AH, BH
	dec temp
	brne Voltage_Measurement__waitloop2

	;add AL, one		;round and divide by 2
	;adc AH, zero
	;lsr AH
	;ror AL

	pop BH
	pop BL
	pop temp
	ret




print_Voltage:
	push AL
	push AH

	;set voltage divider, ADC-reference for voltage measurement
	sbi DDRA, PIN_VCELL_SWITCH	;enable voltage divider
	cbi PORTA, PIN_VCELL_SWITCH	;disable pull-up-resistor

	ldi AL, ADMUX_voltage_measure_AREF
	rcall get_voltage

	;83, xl,xl,xl,xh,xh,xh,13,10 == data, EOL
	push AL
	ldi AL, 83
	rcall putchar
	mov AL, AH
	rcall putchar
	rcall putchar
	rcall putchar
	pop AL
	push AL
	rcall putchar
	rcall putchar
	rcall putchar
	; EOL
	ldi AL, 13	; "\r"
	rcall putchar
	ldi AL, 10	; "\n"
	rcall putchar
	pop AL

	pop AH
	pop AL
	ret



;***********************************************************
;* wait_until_button_pressed                               *
;***********************************************************
.DEF on_off_counter = r16
wait_until_button_pressed:	;here, wait for negative edge
	push on_off_counter

off_A:
	clr on_off_counter
off_B:
	rcall wait_100us
	sbic PINB, PIN_BUTTON
	rjmp on_A
	inc on_off_counter
	brvs off_C
	rjmp off_B

off_C:	;now pin was off for more than 256*100us = 25.6 ms
	sbis PINB, PIN_BUTTON
	rjmp off_C

on_A:
	clr on_off_counter
on_B:
	rcall wait_100us
	sbis PINB, PIN_BUTTON
	rjmp off_A
	inc on_off_counter
	brvs on_C
	rjmp on_B

on_C: ;now pin was on for more than 256*100us = 25.6 ms
	sbic PINB, PIN_BUTTON
	rjmp on_C

	pop on_off_counter
	ret



;***********************************************************
;* plot16                                                  *
;***********************************************************
plot16:
.DEF x_lo = r16
.DEF x_hi = r17
.DEF i = r18
push i
push x_lo
push x_hi
push ZL	; low byte of Z-Pointer
in ZL, SPL
ldd x_hi, Z+7
ldd x_lo, Z+8

	ldi i, 16
plot16_start:
	nop					; 13
	nop					; 14
	nop					; 15
	nop					; 16
	nop					; 17
	nop					; 18
	nop					; 19
	nop					; 20
	nop					; 21
	nop					; 22
	rol x_lo			; 1
	rol x_hi			; 2
	brcc plot16_bit_is_low		; 3
	cbi PLOT16_PORT, PLOT16_PIN		; 4,5 Impulsmarkierung
	sbi PLOT16_PORT, PLOT16_PIN		; 6,7
	dec i				; 8
	nop					; 9
	nop					; 10
	brne plot16_start	; 11,12
	rjmp plot16_weiter; 12,13
plot16_bit_is_low:
	nop					; 5
	sbi PLOT16_PORT, PLOT16_PIN	; 6,7 Impulsmarkierung
	cbi PLOT16_PORT, PLOT16_PIN		; 8,9
	dec i				; 10
	brne plot16_start	; 11, 12
	nop					; 12
	nop					; 13
plot16_weiter:
	nop					; 14
	nop					; 15
	nop					; 16
	nop					; 17
	nop					; 18
	nop					; 19
	nop					; 20
	nop					; 21
	nop					; 22
	nop					; 1
	nop					; 2
	nop					; 3
	cbi PLOT16_PORT, PLOT16_PIN		; 4,5 Impulsmarkierung
	sbi PLOT16_PORT, PLOT16_PIN		; 6,7
	cbi PLOT16_PORT, PLOT16_PIN		; 8,9

pop ZL
pop x_hi
pop x_lo
pop i
ret



;***************************************************************************
;*
;* "putchar"
;*
;* This subroutine transmits the byte stored in the "Txbyte" register
;* The number of stop bits used is set with the sb constant
;*
;* Number of words	:14 including return
;* Number of cycles	:Depens on bit rate
;* Low registers used	:None
;* High registers used	:2 (bitcnt,Txbyte)
;* Pointers used	:None
;*
;***************************************************************************
.equ		sb	=1		;Number of stop bits (1, 2, ...)
.equ		b = 233		;9600bps @ 16MHz
.DEF bitcnt = r19
.DEF tmp = r20

putchar:	
	push bitcnt
	push AL	; byte to transmit
	push tmp

		ldi	bitcnt,9+sb	;1+8+sb (sb is # of stop bits)
		com	AL		;Invert everything
		sec			;Start bit

putchar0:	
		brcc	putchar1	;If carry set
		sbi	TXD_PORT,TXD_PIN	;    send a '0'
		rjmp	putchar2	;else	

putchar1:
		cbi	TXD_PORT,TXD_PIN	;    send a '1'
		nop

putchar2:
UART_delay:	
		ldi	tmp,b
UART_delay1:
		nop					;1
		nop					;2
		nop					;3
		nop					;4
		dec	tmp				;5
		brne	UART_delay1	;6,7

		lsr	AL		;Get next bit
		dec	bitcnt		;If not all bit sent
		brne	putchar0	;   send next
					;else
	pop tmp
	pop AL
	pop bitcnt
	ret			;return












;***************************************************************************
;*
;* "mpy16u" - 16x16 Bit Unsigned Multiplication
;*
;* This subroutine multiplies the two 16-bit register variables 
;* mp16uH:mp16uL and mc16uH:mc16uL.
;* The result is placed in m16u3:m16u2:m16u1:m16u0.
;*  
;* Number of words	:14 + return
;* Number of cycles	:153 + return
;* Low registers used	:None
;* High registers used  :7 (mp16uL,mp16uH,mc16uL/m16u0,mc16uH/m16u1,m16u2,
;*                          m16u3,mcnt16u)	
;*
;***************************************************************************
;***** Subroutine Register Variables
.def	mc16uL	=r16		;multiplicand low byte
.def	mc16uH	=r17		;multiplicand high byte
.def	mp16uL	=r18		;multiplier low byte
.def	mp16uH	=r19		;multiplier high byte
.def	m16u0	=r18		;result byte 0 (LSB)
.def	m16u1	=r19		;result byte 1
.def	m16u2	=r20		;result byte 2
.def	m16u3	=r21		;result byte 3 (MSB)
.def	mcnt16u	=r26		;loop counter
;***** Code
mpy16u:	
	push r26
	clr	m16u3		;clear 2 highest bytes of result
	clr	m16u2
	ldi	mcnt16u,16	;init loop counter
	lsr	mp16uH
	ror	mp16uL
m16u_1:	brcc	noad8		;if bit 0 of multiplier set
	add	m16u2,mc16uL	;add multiplicand Low to byte 2 of res
	adc	m16u3,mc16uH	;add multiplicand high to byte 3 of res
noad8:	ror	m16u3		;shift right result byte 3
	ror	m16u2		;rotate right result byte 2
	ror	m16u1		;rotate result byte 1 and multiplier High
	ror	m16u0		;rotate result byte 0 and multiplier Low
	dec	mcnt16u		;decrement loop counter
	brne	m16u_1		;if not done, loop more
	pop r26
	ret


;***************************************************************************
;*
;* "mpy16s" - 16x16 Bit Signed Multiplication
;*
;* This subroutine multiplies signed the two 16-bit register variables 
;* mp16sH:mp16sL and mc16sH:mc16sL.
;* The result is placed in m16s3:m16s2:m16s1:m16s0.
;* The routine is an implementation of Booth's algorithm. If all 32 bits
;* in the result are needed, avoid calling the routine with
;* -32768 ($8000) as multiplicand
;*  
;* Number of words	:16 + return
;* Number of cycles	:210/226 (Min/Max) + return
;* Low registers used	:None
;* High registers used  :7 (mp16sL,mp16sH,mc16sL/m16s0,mc16sH/m16s1,
;*			    m16s2,m16s3,mcnt16s)	
;*
;***************************************************************************

;***** Subroutine Register Variables

.def	mc16sL	=r16		;multiplicand low byte
.def	mc16sH	=r17		;multiplicand high byte
.def	mp16sL	=r18		;multiplier low byte
.def	mp16sH	=r19		;multiplier high byte
.def	m16s0	=r18		;result byte 0 (LSB)
.def	m16s1	=r19		;result byte 1
.def	m16s2	=r20		;result byte 2
.def	m16s3	=r21		;result byte 3 (MSB)
.def	mcnt16s	=r26		;loop counter

;***** Code
mpy16s:	
	push r26
	clr	m16s3		;clear result byte 3
	sub	m16s2,m16s2	;clear result byte 2 and carry
	ldi	mcnt16s,16	;init loop counter
m16s_1:	brcc	m16s_2		;if carry (previous bit) set
	add	m16s2,mc16sL	;    add multiplicand Low to result byte 2
	adc	m16s3,mc16sH	;    add multiplicand High to result byte 3
m16s_2:	sbrc	mp16sL,0	;if current bit set
	sub	m16s2,mc16sL	;    sub multiplicand Low from result byte 2
	sbrc	mp16sL,0	;if current bit set
	sbc	m16s3,mc16sH	;    sub multiplicand High from result byte 3
	asr	m16s3		;shift right result and multiplier
	ror	m16s2
	ror	m16s1
	ror	m16s0
	dec	mcnt16s		;decrement counter
	brne	m16s_1		;if not done, loop more	
	pop r26
	ret



;***************************************************************************
;*
;* "div16u" - 16/16 Bit Unsigned Division
;*
;* This subroutine divides the two 16-bit numbers 
;* "dd8uH:dd8uL" (dividend) and "dv16uH:dv16uL" (divisor). 
;* The result is placed in "dres16uH:dres16uL" and the remainder in
;* "drem16uH:drem16uL".
;*  
;* Number of words	:19
;* Number of cycles	:235/251 (Min/Max)
;* Low registers used	:2 (drem16uL,drem16uH)
;* High registers used  :5 (dres16uL/dd16uL,dres16uH/dd16uH,dv16uL,dv16uH,
;*			    dcnt16u)
;*
;***************************************************************************

;***** Subroutine Register Variables

.def	drem16uL=r20
.def	drem16uH=r21
.def	dres16uL=r16
.def	dres16uH=r17
.def	dd16uL	=r16
.def	dd16uH	=r17
.def	dv16uL	=r18
.def	dv16uH	=r19
.def	dcnt16u	=r26

;***** Code

div16u:	
	push dcnt16u
	clr	drem16uL	;clear remainder Low byte
	sub	drem16uH,drem16uH;clear remainder High byte and carry
	ldi	dcnt16u,17	;init loop counter
d16u_1:	rol	dd16uL		;shift left dividend
	rol	dd16uH
	dec	dcnt16u		;decrement counter
	brne	d16u_2		;if done
	pop dcnt16u
	ret			;    return
d16u_2:	rol	drem16uL	;shift dividend into remainder
	rol	drem16uH
	sub	drem16uL,dv16uL	;remainder = remainder - divisor
	sbc	drem16uH,dv16uH	;
	brcc	d16u_3		;if result negative
	add	drem16uL,dv16uL	;    restore remainder
	adc	drem16uH,dv16uH
	clc			;    clear carry to be shifted into result
	rjmp	d16u_1		;else
d16u_3:	sec			;    set carry to be shifted into result
	rjmp	d16u_1
	

